"""
Tests.

This package regroups all common helpers to work with a test cases.
"""

from .utils import generate_random_email, generate_random_string

__all__ = [
    "generate_random_email",
    "generate_random_string",
]
